import Canvas from '../references/canvas.js';

import './project.scss';


class GanttChart {
  constructor(el, options) {
    const a = this;

    a.Canvas = new Canvas(el);
    a.Options = Object.assign(a.DefaultOptions, options);

    a.Debug = false;
    a.Tasks = [];
    a.StartDate = null;
    a.Duration = 30;
    a.StartOfWeek = 1;

    a.#initialiseComponents();
  }

  #initialiseComponents() {
    const a = this;

    if (!a.Canvas.el.classList.contains("literyzjs-project")) {
      a.Canvas.el.classList.add("literyzjs-project");
    }

    if (a.Canvas.FlowContainer.classList.contains("border")) {
      a.Canvas.FlowContainer.classList.remove("border");
    }

    if (!a.Canvas.FlowContainer.classList.contains("gantt-chart")) {
      a.Canvas.FlowContainer.classList.add("gantt-chart");
    }

  }


  get DefaultOptions() {
    return {
      DayWidth: 24,
      HeaderRow: {
        Height: [ 20, 21 ]
      },
      Row: {
        Height: 28,
        BackColour: "rgba(235, 235, 235, 0.3)",
        Task: {
          Height: 14,
          BorderColour: "#555555",
          FillColour: "#D8EEDB"
        },
        OrphanTask: {
          Height: 14,
          BorderColour: "#555555",
          FillColour: "#9CC2E6"
        },
        ChildTask: {
          Height: 14,
          BorderColour: "#555555",
          FillColour: "#FFF5C1"
        },
        CollatedTask: {
          Height: 6,
          BorderColour: "#555555",
          FillColour: "#555555"
        }
      },
      Column: {
        SatColour: "rgba(233, 237, 239, 0.8)",
        SunColour: "rgba(233, 237, 239, 0.9)",
      },
      Line: {
        Colour: "#555555",
        Width: 1,
        ArrowSize: 5
      },
      DateFont: "7pt sans-serif",
      DateForeColour: "#636363",
      BorderWidth: 1,
      BorderColour: "#B8B8B8",
      BorderDashPattern: [1, 1],
      MinimumRowCount: 6,
      ShowDateLines: true,
      ShowStartDayOfWeekLine: true,
      ShowRowLines: true,
      ShowRowStripes: true
    };
  }

  get HeaderRow1Rectangle() {
    return {
      X: 0,
      Y: 0,
      W: this.Canvas.Width,
      H: (this.Options.HeaderRow.Height[0] - this.Options.BorderWidth)
    };
  }

  get HeaderRow2Rectangle() {
    const a = this;

    return {
      X: 0,
      Y: this.Options.HeaderRow.Height[0],
      W: this.Canvas.Width,
      H: (this.Options.HeaderRow.Height[1] - this.Options.BorderWidth)
    };
  }

  get HeaderHeight() {
    const a = this;

    return a.Options.HeaderRow.Height[0] + a.Options.HeaderRow.Height[1];
  }


  Clear() {
    const a = this;

    a.Canvas.Clear();

    a.StartDate = new Date();
    a.Duration = 30;
    a.StartOfWeek = 1;    // 1 = Monday
    a.Tasks = [];

    a.Invalidate();
  }

  Invalidate() {
    const a = this;

    a.Canvas.Clear();

    const width = ((a.Duration + 2) * a.Options.DayWidth);
    const height = (Math.max(a.Tasks.length, a.Options.MinimumRowCount) * a.Options.Row.Height) + a.HeaderHeight;

    a.Canvas.Size = { W: width, H: height };

    a.Canvas.Invalidate();

    a.#drawChartHeader();
    a.#drawColumnLayout();
    a.#drawRowLayout();
    a.#drawTasks();
    a.#drawConnectorLines();
  }

  Load(project) {
    const a = this;

    a.Canvas.Clear();

    if (project == null) {
      return;
    }

    a.StartDate = new Date(project.StartDate);
    a.Duration = project.Duration;
    a.StartOfWeek = project.Project.StartOfWeek;
    a.Tasks = project.ExportTasks();

    a.Invalidate();
  }

  #drawChartHeader() {
    const a = this;

    const displayDays = a.Duration + 2;

    if (a.Debug) {
      a.Canvas.DrawRectangle(a.HeaderRow1Rectangle, "red", {});
      a.Canvas.DrawRectangle(a.HeaderRow2Rectangle, "orange", {});
    }

    let startDate = new Date(a.StartDate);
    startDate.addDays(-1);

    // Draw horizontal line under months
    a.#drawHorizontalLine(a.Options.HeaderRow.Height[0]);

    // Draw vertical lines for dates
    for (let i=1; i<displayDays; i++) {
      a.Canvas.DrawVerticalLine((a.Options.DayWidth * i), (a.Options.HeaderRow.Height[0] + a.Options.BorderWidth), (a.Options.HeaderRow.Height[1] - a.Options.BorderWidth), a.Options.BorderColour, {});
    }

    // Draw horizontal line under dates
    a.#drawHorizontalLine(a.HeaderHeight);

    // Write dates
    for (let i=0; i<displayDays; i++) {
      const date = Date.addDays(startDate, i);
      const x = (a.Options.DayWidth * i);

      // Draw month
      if (date.getDate() == 1) {
        const size = a.Canvas.MeasureText(a.Options.DateFont, "#");
        const monthPoint = {
          X: (x + 2),
          Y: Math.half(a.Options.HeaderRow.Height[0] - size.H)
        };

        a.Canvas.DrawText(monthPoint.X, monthPoint.Y, date.toCString("MMMM"), a.Options.DateFont, a.Options.DateForeColour);
      }

      // Draw day
      const dateRectangle = {
        X: x,
        Y: a.Options.HeaderRow.Height[0],
        W: a.Options.DayWidth - a.Options.BorderWidth,
        H: (a.Options.HeaderRow.Height[1] - a.Options.BorderWidth)
      };

      a.Canvas.FillText(dateRectangle, date.getDate(), a.Options.DateFont, a.Options.DateForeColour);

      if (a.Debug) {
        a.Canvas.DrawRectangle(dateRectangle, "red", {});
      }

    }
  }

  #drawColumnLayout() {
    const a = this;

    const height = a.Canvas.Height;
    const displayDays = a.Duration + 2;

    let startDate = new Date(a.StartDate);
    startDate.addDays(-1);

    // Write dates
    for (let i=0; i<displayDays; i++) {
      const date = Date.addDays(startDate, i);
      const x = (a.Options.DayWidth * i);

      const dateRectangle = {
        X: x,
        Y: a.HeaderHeight,
        W: a.Options.DayWidth - a.Options.BorderWidth,
        H: height - a.Options.BorderWidth
      };

      // Fill background for Saturday and Sunday
      if (date.getDay() == 6) {
        a.Canvas.DrawRectangle(dateRectangle, a.Options.Column.SatColour, { FillColour: a.Options.Column.SatColour });
      } else if (date.getDay() == 0) {
        a.Canvas.DrawRectangle(dateRectangle, a.Options.Column.SunColour, { FillColour: a.Options.Column.SunColour });
      }

      // Draw vertical date lines
      if (a.StartOfWeek == date.getDay()) {
        if (a.Options.ShowStartDayOfWeekLine) {
          a.Canvas.DrawVerticalLine(dateRectangle.X, a.HeaderHeight, dateRectangle.H, a.Options.BorderColour, {});
        }
      } else {
        if (a.Options.ShowDateLines) {
          a.Canvas.DrawVerticalLine(dateRectangle.X, a.HeaderHeight, dateRectangle.H, a.Options.BorderColour, { LineDash: a.Options.BorderDashPattern });
        }
      }

    }
  }

  #drawRowLayout() {
    const a = this;

    for (let i=0; i<a.Tasks.length; i++) {
      const rowRectangle = {
        X: 0,
        Y: (a.HeaderHeight + (a.Options.Row.Height * i)),
        W: a.Canvas.Width,
        H: (a.Options.Row.Height - a.Options.BorderWidth)
      };

      if (a.Options.ShowRowStripes) {
        if ((i % 2) == 1) {
          a.Canvas.DrawRectangle(rowRectangle, a.Options.Row.BackColour, { FillColour: a.Options.Row.BackColour });
        }
      }

      if (a.Options.ShowRowLines) {
        a.Canvas.DrawHorizontalLine(0, (rowRectangle.Y + a.Options.Row.Height), rowRectangle.W, a.Options.BorderColour, { LineDash: a.Options.BorderDashPattern });
      }

    }
  }

  #drawTasks() {
    const a = this;

    for (let i=0; i<a.Tasks.length; i++) {
      let style = a.Options.Row.Task;

      if (a.Tasks[i].IsCollated == true) {
        style = a.Options.Row.CollatedTask;
      } else if (a.Tasks[i].CollatedTaskID != null) {
        style = a.Options.Row.ChildTask;
      } else if (a.Tasks[i].PredecessorTaskNo == null) {
        style = a.Options.Row.OrphanTask;
      }

      const rectangle = a.#getTaskRectangle(a.Tasks[i], style);

      if (a.Tasks[i].Duration <= 0) {
        a.Canvas.DrawDiamond(rectangle, style.BorderColour, { FillColour: style.FillColour });
      } else {
        a.Canvas.DrawRectangle(rectangle, style.BorderColour, { FillColour: style.FillColour });
      }
    }
  }

  #drawConnectorLines() {
    const a = this;

    for (let i=0; i<a.Tasks.length; i++) {
      if (a.Tasks[i].PredecessorTaskID == null) {
        continue;
      }

      if (a.Tasks[i].PredecessorTaskNo == null) {
        continue;
      }

      const style = ((a.Tasks[i].IsCollated == true) ? a.Options.Row.CollatedTask : a.Options.Row.Task);
      const rectangle = a.#getTaskRectangle(a.Tasks[i], style);
      const paddingX = (a.Options.Row.CollatedTask.Height + a.Options.BorderWidth);
      const offsetX = (rectangle.X + paddingX);

      const predecessorTask = a.Tasks.first("Order", a.Tasks[i].PredecessorTaskNo);
      if (predecessorTask == null) {
        continue;
      }

      const predecessorStyle = ((predecessorTask.IsCollated == true) ? a.Options.Row.CollatedTask : a.Options.Row.Task);
      const predecessorRectangle = a.#getTaskRectangle(predecessorTask, predecessorStyle);

      let points = [];
      points.push({ X: (predecessorRectangle.X + predecessorRectangle.W), Y: (predecessorRectangle.Y + Math.half(predecessorRectangle.H)) });
      points.push({ X: offsetX, Y: (predecessorRectangle.Y + Math.half(predecessorRectangle.H)) });

      points.push({ X: offsetX, Y: (rectangle.Y - paddingX) });
      points.push({ X: offsetX, Y: rectangle.Y });

      a.Canvas.DrawLines(points, a.Options.Line.Colour, { LineWidth: a.Options.Line.Width });

      const arrowRectangle = {
        X: (offsetX - Math.half(a.Options.Line.ArrowSize)),
        Y: (rectangle.Y - a.Options.Line.ArrowSize),
        W: a.Options.Line.ArrowSize,
        H: a.Options.Line.ArrowSize
      }

      a.Canvas.DrawArrowS(arrowRectangle, a.Options.Line.Colour, { FillColour: a.Options.Line.Colour });
    }
  }

  #drawHorizontalLine(y) {
    const a = this;
    const width = a.Canvas.Width;

    a.Canvas.DrawHorizontalLine(0, y, width, this.Options.BorderColour);
  }

  #getTaskRectangle(task, style) {
    const a = this;
    const i = (task.Order - 1);

    const rowRectangle = {
      X: 0,
      Y: (a.HeaderHeight + (a.Options.Row.Height * i)),
      W: a.Canvas.Width,
      H: (a.Options.Row.Height - a.Options.BorderWidth)
    };

    if (a.Debug) {
      a.Canvas.DrawRectangle(rowRectangle, "red", {});
    }

    if (task.Duration <= 0) {
      let result = {
        X: (a.Options.DayWidth * (Date.diffDays(a.StartDate, task.StartDate) + 1)),
        Y: (a.HeaderHeight + (a.Options.Row.Height * i) + Math.half(rowRectangle.H - style.Height)),
        W: style.Height,
        H: style.Height
      };

      result.X += Math.half(a.Options.DayWidth - (style.Height + a.Options.BorderWidth));

      return result;
    } else {
      return {
        X: (a.Options.DayWidth * (Date.diffDays(a.StartDate, task.StartDate) + 1)),
        Y: (a.HeaderHeight + (a.Options.Row.Height * i) + Math.half(rowRectangle.H - style.Height)),
        W: (a.Options.DayWidth * task.Duration),
        H: style.Height
      };
    }
  }

}


export default GanttChart;